unit Player;

interface

uses Windows, MMSystem, ActiveX, DShowSub, EVRSub;

type
  TFilterInfo = record
    Name: string;
    Filter: IBaseFilter;
    Prop: ISpecifyPropertyPages;
  end;

  TMediaTypeInfo = record
    AudioSampleRate: Longint;
    AudioBitRate: Longint;
    AudioChannels: Longint;
    AudioFormatTag: Longint;
    VideoWidth: Longint;
    VideoHeight: Longint;
    VideoFrameRate: Double;
    VideoCodec: string;
  end;

  TPlayerState = (psUnknown, psStopped, psPaused, psPlaying);
  TPlayerRenderer = (prDefault, prVMR9, prEVR);
  TPlayerSourceType = (pstNone, pstFile, pstDvd, pstUrl);

  TPlayer = class(TObject)
  private
    BasicAudio: IBasicAudio;
    Duration: LONGLONG;
    DvdControl2: IDvdControl2;
    DvdInfo2: IDvdInfo2;
    DvdNavFilter: IBaseFilter;
    EVRFilter: IBaseFilter;
    EVRWnd: HWND;
    GraphBuilder: IGraphBuilder;
    //Line21Decoder: IAMLine21Decoder;
    MFVideoDisplayControl: IMFVideoDisplayControl;
    MediaControl: IMediaControl;
    MediaSeeking: IMediaSeeking;
    Parent: HWND;
    VMR9Filter: IBaseFilter;
    VideoWindow: IVideoWindow;
    WMAsfReader: IBaseFilter;
    procedure FreeInterfaces;
    procedure GetFilterInfo;
    procedure AddVideoRenderer;
    procedure LoadByWMAsfReader(const Path: string);
    function GetImageMF(var Buf: PChar): Cardinal;
    function GetImageBasicVideo(var Buf: PChar): Cardinal;
    function GetImageSampleGrabber(var Buf: PChar): Cardinal;
  public
    DvdChapter: Integer;
    DvdChapterCount: ULONG;
    DvdTitle: Integer;
    DvdTitleCount: ULONG;
    FilterInfo: array of TFilterInfo;
    HasVideo: Boolean;
    Height: Longint;
    Len: Longint;
    Path: string;
    Renderer: TPlayerRenderer;
    SourceType: TPlayerSourceType;
    State: TPlayerState;
    VideoAspect: Double;
    VideoHeight: Longint;
    VideoWidth: Longint;
    WarnMsg: string;
    Width: Longint;
    MediaTypeInfo: TMediaTypeInfo;
    constructor Create(Parent: HWND);
    destructor Destroy; override;
    function Capture(var Buf: PChar): Cardinal;
    //function ClosedCaption: Boolean;
    function DvdGetAudioLanguage(Num: Integer): string;
    function DvdGetSubpictureLanguage(Num: Integer): string;
    function DvdIsAudioEnabled(Num: Integer): BOOL;
    function DvdIsSubpictureEnabled(Num: Integer): BOOL;
    function Loaded: Boolean;
    function Pos: Longint;
    function Seekable: Boolean;
    procedure DvdGetAudio(var Count, Current: ULONG);
    procedure DvdGetSubpicture(var Count, Current: ULONG;
        var Display: Boolean);
    procedure DvdPlayChapter(Chapter: Integer);
    procedure DvdPlayTitle(Title: Integer);
    procedure DvdSetAudio(Num: Integer);
    procedure DvdSetSubpicture(Num: Integer);
    procedure DvdShowSubpicture(Flag: Boolean);
    procedure Init;
    procedure LoadFile(const Path: string);
    procedure OpenDvd(const Path: string);
    procedure Pause;
    procedure Play;
    procedure Seek(MilliSec: Integer);
    procedure SeekTo(MilliSec: Integer);
    procedure SetBounds(Left, Top, Width, Height: Longint);
    //procedure SetClosedCaption(Flag: Boolean);
    procedure SetVolume(Value: Integer);
    procedure ShowFilterProperty(Index: Integer);
    procedure Stop;
  end;

  EDShowError = class(TObject)
    Msg: string;
    constructor Create(Code: HRESULT);
  end;

  function IsUrl(const S: string): Boolean;

implementation

uses NLib;

function IsUrl(const S: string): Boolean;
begin
  Result := (Copy(S, 1, 7) = 'http://') or (Copy(S, 1, 8) = 'https://') or
            (Copy(S, 1, 7) = 'rtsp://') or (Copy(S, 1, 6) = 'mms://');
end;

function DShowError(Code: HRESULT): string;
const
  INET_E_RESOURCE_NOT_FOUND = $800c0005;
  INET_E_OBJECT_NOT_FOUND = $800c0006;
  INET_E_DOWNLOAD_FAILURE = $800c0008;
var
  buf: array[0..MAX_ERROR_TEXT_LEN-1] of Char;
begin
  SetString(Result, buf, AMGetErrorText(Code, buf, SizeOf(buf)));
  if Result <> '' then
    Exit;

  case Cardinal(Code) of
    // URL moniker errors
    INET_E_RESOURCE_NOT_FOUND: Result := 'Resource not found';
    INET_E_OBJECT_NOT_FOUND: Result := 'Object not found';
    INET_E_DOWNLOAD_FAILURE: Result := 'Download failure';
  else
    SetString(Result, buf,
      wvsprintf(buf, 'Unknown DirectShow error: 0x%2x', @code));
  end;
end;

procedure CheckResult(Code: HRESULT);
begin
  if FAILED(Code) then
    raise EDShowError.Create(Code);
end;

function DvdTimeCodeToMilliSec(const TimeCode: DVD_HMSF_TIMECODE): Integer;
begin
  Result := (TimeCode.bHours * 3600 + TimeCode.bMinutes * 60 +
    TimeCode.bSeconds) * 1000;
end;

function DvdTimeCode(MilliSec: Integer): DVD_HMSF_TIMECODE;
begin
  Result.bHours := MilliSec div 1000 div 3600;
  Result.bMinutes := MilliSec div 1000 mod 3600 div 60;
  Result.bSeconds := MilliSec div 1000 mod 60;
  Result.bFrames := 0;
end;

{ Exception }

constructor EDShowError.Create;
begin
  Msg := DShowError(Code);
end;


{ TPlayer }

constructor TPlayer.Create(Parent: HWND);
begin
  Self.Parent := Parent;
  EVRWnd := CreateWindow('STATIC', 'NkVp EVR Window', WS_CHILD or SS_OWNERDRAW,
      0, 0, 0, 0, Parent, 0, 0, nil);
end;

destructor TPlayer.Destroy;
begin
  FreeInterfaces;
  DestroyWindow(EVRWnd);
end;

procedure TPlayer.FreeInterfaces;
var
  i: Integer;
begin
  for i := 0 to High(FilterInfo) do
    with FilterInfo[i] do
    begin
      Filter := nil;
      Prop := nil;
    end;
  BasicAudio := nil;
  DvdControl2 := nil;
  DvdInfo2 := nil;
  DvdNavFilter := nil;
  EVRFilter := nil;
  GraphBuilder := nil;
  //Line21Decoder := nil;
  MFVideoDisplayControl := nil;
  MediaControl := nil;
  MediaSeeking := nil;
  VMR9Filter := nil;
  VideoWindow := nil;
  WMAsfReader := nil;
end;

procedure TPlayer.AddVideoRenderer;
  procedure AddVMR9Filter;
  begin
    if FAILED(CoCreateInstance(CLSID_VideoMixingRenderer9, nil,
        CLSCTX_INPROC_SERVER, IID_IBaseFilter, VMR9Filter)) then
      Exit;
    if FAILED(GraphBuilder.AddFilter(VMR9Filter,
        'Video Mixing Renderer 9')) then
    begin
      VMR9Filter := nil;
      Exit;
    end;
  end;
  procedure AddEVRFilter;
  begin
    if FAILED(CoCreateInstance(CLSID_EnhancedVideoRenderer, nil,
        CLSCTX_INPROC_SERVER, IID_IBaseFilter, EVRFilter)) then
      Exit;
    if FAILED(GraphBuilder.AddFilter(EVRFilter,
        'Enhanced Video Renderer')) then
    begin
      EVRFilter := nil;
      Exit;
    end;
    (EVRFilter as IMFGetService).GetService(MR_VIDEO_RENDER_SERVICE,
        IID_IMFVideoDisplayControl, MFVideoDisplayControl);
    MFVideoDisplayControl.SetVideoWindow(EVRWnd);
  end;
begin
  if Renderer = prEVR then
  begin
    AddEVRFilter;
    if Assigned(EVRFilter) then
      Exit;
    Renderer := prVMR9;
  end;

  if Renderer = prVMR9 then
    AddVMR9Filter;
end;

procedure TPlayer.OpenDvd(const Path:string);
var
  nvol, vol: ULONG;
  side: DVD_DISC_SIDE;
  pin: IPin;
  enum: IEnumPins;
  f: ULONG;
begin
  Stop;
  SourceType := pstNone;
  WarnMsg := '';
  FreeInterfaces;

  CheckResult(CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC_SERVER,
      IID_IGraphBuilder, GraphBuilder));
  CheckResult(CoCreateInstance(CLSID_DVDNavigator, nil, CLSCTX_INPROC_SERVER,
      IID_IBaseFilter, DVDNavFilter));
  CheckResult(GraphBuilder.AddFilter(DVDNavFilter, 'DVD Navigator'));

  {
  CheckResult(CoCreateInstance(CLSID_Line21Decoder, nil, CLSCTX_INPROC_SERVER,
      IID_IAMLine21Decoder, Line21Decoder));
  CheckResult(GraphBuilder.AddFilter(Line21Decoder as IBaseFilter,
      'Line 21 Decoder 2'));
  }

  AddVideoRenderer;

  DvdNavFilter.EnumPins(enum);
  while enum.Next(1, pin, f) = S_OK do
    GraphBuilder.Render(pin);

  DvdControl2 := DvdNavFilter as IDvdControl2;
  DvdInfo2 := DvdNavFilter as IDvdInfo2;

  CheckResult(DvdControl2.SetDVDDirectory(PWideChar(WideString(Path))));

  Self.Path := Path;
  DvdInfo2.GetDVDVolumeInfo(nvol, vol, side, DvdTitleCount);
  DvdControl2.SetOption(DVD_ResetOnStop, False);
  SourceType := pstDvd;
  Init;
  Play; // start graph
  DvdPlayTitle(1);
end;

procedure TPlayer.DvdSetAudio;
var
  dvdcmd: IDvdCmd;
begin
  DvdControl2.SelectAudioStream(Num, DVD_CMD_FLAG_Flush, dvdcmd);
end;

procedure TPlayer.DvdGetAudio;
begin
  DvdInfo2.GetCurrentAudio(Count, Current);
end;

function TPlayer.DvdIsAudioEnabled;
begin
  DvdInfo2.IsAudioStreamEnabled(Num, Result)
end;

function TPlayer.DvdGetAudioLanguage;
var
  Lang: LCID;
begin
  if FAILED(DvdInfo2.GetAudioLanguage(Num, Lang)) then
  begin
    Result := '';
    Exit;
  end;
  // -1: for null character
  SetString(Result, nil,
    GetLocaleInfo(Lang, LOCALE_SENGLANGUAGE, nil, 0)-1);
  GetLocaleInfo(Lang, LOCALE_SENGLANGUAGE, PChar(Result), Length(Result));
end;

procedure TPlayer.DvdSetSubpicture;
var
  dvdcmd: IDvdCmd;
begin
  DvdControl2.SelectSubpictureStream(Num, DVD_CMD_FLAG_Flush, dvdcmd);
end;

procedure TPlayer.DvdGetSubpicture;
var
  disabled: BOOL;
begin
  DvdInfo2.GetCurrentSubpicture(Count, Current, disabled);
  Display := not disabled;
end;

procedure TPlayer.DvdShowSubpicture;
var
  dvdcmd: IDvdCmd;
begin
  DvdControl2.SetSubpictureState(Flag, DVD_CMD_FLAG_Flush, dvdcmd);
end;

function TPlayer.DvdIsSubpictureEnabled;
begin
  DvdInfo2.IsSubpictureStreamEnabled(Num, Result)
end;

function TPlayer.DvdGetSubpictureLanguage;
var
  Lang: LCID;
begin
  if FAILED(DvdInfo2.GetSubpictureLanguage(Num, Lang)) then
  begin
    Result := '';
    Exit;
  end;
  // -1: for null character
  SetString(Result, nil,
    GetLocaleInfo(Lang, LOCALE_SENGLANGUAGE, nil, 0)-1);
  GetLocaleInfo(Lang, LOCALE_SENGLANGUAGE, PChar(Result), Length(Result));
end;

procedure TPlayer.DvdPlayTitle;
var
  tc: DVD_HMSF_TIMECODE;
  tcf: ULONG;
  dvdcmd: IDvdCmd;
begin
  if (Title < 1) or (ULONG(Title) > DvdTitleCount) then
    Exit;
  //DvdControl2.Stop;
  if FAILED(DvdControl2.PlayTitle(Title, DVD_CMD_FLAG_Flush, dvdcmd)) then
    Exit;
  DvdTitle := Title;
  if SUCCEEDED(DvdInfo2.GetNumberOfChapters(DvdTitle, DvdChapterCount)) then
    DvdChapter := 1
  else begin
    DvdChapter := 0;
    DvdChapterCount := 0;
  end;
  DvdInfo2.GetTotalTitleTime(tc, tcf);
  Len := DvdTimeCodeToMilliSec(tc);
  Duration := LONGLONG(Len) * 10000;
  Play;
end;

procedure TPlayer.DvdPlayChapter;
var
  dvdcmd: IDvdCmd;
begin
  if SUCCEEDED(DvdControl2.PlayChapter(
      Chapter, DVD_CMD_FLAG_Flush, dvdcmd)) then
    DvdChapter := Chapter;
end;

procedure TPlayer.LoadByWMAsfReader(const Path: string);
var
  enum: IEnumPins;
  pin: IPin;
  fetched: ULONG;
  r: HRESULT;
begin
  if FAILED(CoCreateInstance(CLSID_WMAsfReader, nil, CLSCTX_INPROC_SERVER,
      IID_IBaseFilter, WMAsfReader)) then
    Exit;

  if FAILED(GraphBuilder.AddFilter(WMAsfReader, 'WM ASF Reader')) then
  begin
    WMAsfReader := nil;
    Exit;
  end;

  if FAILED((WMAsfReader as IFileSourceFilter).Load(
      PWideChar(WideString(Path)), nil)) then
  begin
    GraphBuilder.RemoveFilter(WMAsfReader);
    WMAsfReader := nil;
    Exit;
  end;

  WarnMsg := '';
  WMAsfReader.EnumPins(enum);
  while enum.Next(1, pin, fetched) = S_OK do
  begin
    r := GraphBuilder.Render(pin);
    if SUCCEEDED(r) then
    begin
      if r <> S_OK then
      begin
        if WarnMsg <> '' then
          WarnMsg := #13#10 + DShowError(r)
        else
          WarnMsg := DShowError(r);
      end;
    end
    else begin
      WMAsfReader := nil;
      Exit;
    end;
  end;
end;

procedure TPlayer.LoadFile(const Path: string);
var
  r: HRESULT;
begin
  Stop;
  SourceType := pstNone;
  FreeInterfaces;

  CheckResult(CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC_SERVER,
      IID_IGraphBuilder, GraphBuilder));

  AddVideoRenderer;

  if Assigned(VMR9Filter) then
    (VMR9Filter as IVMRAspectRatioControl9).SetAspectRatioMode(VMR_ARMODE_NONE);

  if Assigned(EVRFilter) then
    MFVideoDisplayControl.SetAspectRatioMode(MFVideoARMode_None);

  LoadByWMAsfReader(Path);
  if WMAsfReader = nil then
  begin
    WarnMsg := '';
    r := GraphBuilder.RenderFile(PWideChar(WideString(Path)), nil);
    if SUCCEEDED(r) then
    begin
      if r <> S_OK then
        WarnMsg := DShowError(r);
    end
    else begin
      GraphBuilder := nil;
      raise EDShowError.Create(r);
    end;
  end;

  if IsUrl(path) then
    SourceType := pstUrl
  else
    SourceType := pstFile;
  Init;
  Self.Path := Path;
  if SUCCEEDED(MediaSeeking.GetDuration(Duration)) then
    Len := Duration div 10000
  else begin
    Duration := -1;
    Len := -1;
  end;
end;

procedure TPlayer.Init;
var
  x, y: Integer;
  fs: FILTER_STATE;
  vsize, asize: TSize;
  basicvideo2: IBasicVideo2;
begin
  VideoWidth := 0;
  VideoHeight := 0;
  VideoAspect := 0;
  HasVideo := False;

  MediaControl := GraphBuilder as IMediaControl;
  MediaSeeking := GraphBuilder as IMediaSeeking;
  BasicAudio := GraphBuilder as IBasicAudio;
  basicvideo2 := GraphBuilder as IBasicVideo2;

  if Assigned(EVRFilter) then
  begin
    FillChar(vsize, SizeOf(vsize), 0);
    FillChar(asize, SizeOf(asize), 0);
    MFVideoDisplayControl.GetNativeVideoSize(@vsize, @asize);
    VideoWidth := vsize.cx;
    VideoHeight := vsize.cy;
    if (vsize.cy > 0) and (asize.cx > 0) then
       VideoAspect := (vsize.cx * asize.cy) / (vsize.cy * asize.cx);
    HasVideo := VideoWidth > 0;
    if HasVideo then
      ShowWindow(EVRWnd, SW_SHOW)
    else begin
      ShowWindow(EVRWnd, SW_HIDE);
      GraphBuilder.RemoveFilter(EVRFilter);
      MFVideoDisplayControl := nil;
      EVRFilter := nil;
    end;
  end
  else begin
    ShowWindow(EVRWnd, SW_HIDE);
    VideoWindow := GraphBuilder as IVideoWindow;
    VideoWindow.put_AutoShow(OAFALSE);
    VideoWindow.put_Owner(OAHWND(Parent));
    VideoWindow.put_WindowStyle(WS_CHILD);
    VideoWindow.put_MessageDrain(Parent);
    HasVideo := SUCCEEDED(basicvideo2.GetVideoSize(VideoWidth, VideoHeight));
    if not HasVideo and Assigned(VMR9Filter) then
    begin
      GraphBuilder.RemoveFilter(VMR9Filter);
      VMR9Filter := nil;
    end;
  end;

  GetFilterInfo;

  if SourceType = pstFile then
  begin
    MediaControl.StopWhenReady;
    while MediaControl.GetState(10, fs) = VFW_S_STATE_INTERMEDIATE do;
  end;

  if not Assigned(EVRFilter) then
    if SUCCEEDED(basicvideo2.GetPreferredAspectRatio(x, y)) then
      if (x > 0) and (y > 0) then
      begin
        VideoAspect := (VideoWidth * y) / (VideoHeight * x);
        if VideoAspect = 1.125 then
          VideoAspect := 1.1; // NTSC
      end;

  if Assigned(VideoWindow) then
    VideoWindow.put_Visible(OATRUE);
end;

procedure TPlayer.Play;
begin
  if Assigned(MediaControl) then
  begin
    MediaControl.Run;
    State := psPlaying;
  end;
end;

procedure TPlayer.Pause;
begin
  if Assigned(MediaControl) then
  begin
    MediaControl.Pause;
    State := psPaused;
  end;
end;

procedure TPlayer.Stop;
begin
  if Assigned(MediaControl) then
  begin
    MediaControl.Stop;
    State := psStopped;
  end;
end;

procedure TPlayer.SetVolume(Value: Integer);
begin
  if Assigned(BasicAudio) then
    BasicAudio.put_Volume(Value);
end;

function TPlayer.Loaded: Boolean;
begin
  Result := SourceType <> pstNone;
end;

function TPlayer.Pos;
var
  p: LONGLONG;
  loc: DVD_PLAYBACK_LOCATION2;
  r: HRESULT;
begin
  if SourceType = pstDvd then
  begin
    r := DvdInfo2.GetCurrentLocation(loc);
    if SUCCEEDED(r) then
    begin
      DvdTitle := loc.TitleNum;
      DvdChapter := loc.ChapterNum;
      Result := DvdTimeCodeToMilliSec(loc.TimeCode);
    end else
      Result := 0;
  end
  else if Assigned(MediaSeeking) then
  begin
    MediaSeeking.GetCurrentPosition(p);
    Result := p div 10000;
  end else
    Result := 0;
end;

procedure TPlayer.SetBounds(Left, Top, Width, Height: Longint);
var
  nrc: TMFVideoNormalizedRect;
  rc: TRect;
begin
  if Assigned(VideoWindow) then
  begin
    Self.Width := Width;
    Self.Height := Height;
    VideoWindow.SetWindowPosition(Left, Top, Width, Height);
  end
  else if Assigned(MFVideoDisplayControl) then
  begin
    Self.Width := Width;
    Self.Height := Height;
    nrc.left := 0;
    nrc.top := 0;
    nrc.right := 1;
    nrc.bottom := 1;
    rc.Left := 0;
    rc.Top := 0;
    rc.Right := Width;
    rc.Bottom := Height;
    MFVideoDisplayControl.SetVideoPosition(@nrc, @rc);
    MoveWindow(EVRWnd, Left, Top, Width, Height, False);
    MFVideoDisplayControl.RepaintVideo;
  end;
end;

procedure TPlayer.Seek(MilliSec: Integer);
begin
  SeekTo(Pos + MilliSec);
end;

procedure TPlayer.SeekTo(MilliSec: Integer);
var
  p, pp: LONGLONG;
  dvdcmd: IDvdCmd;
  tc: DVD_HMSF_TIMECODE;
begin
  if not Seekable then
    Exit;
  p := LONGLONG(MilliSec) * 10000;
  if p < 0 then
    p := 0;
  if p > Duration then
    p := Duration;
  if SourceType = pstDvd then
  begin
    tc := DvdTimeCode(p div 10000);
    DvdControl2.PlayAtTime(tc, DVD_CMD_FLAG_Flush, dvdcmd);
  end
  else if Assigned(MediaSeeking) then
    MediaSeeking.SetPositions(p, AM_SEEKING_AbsolutePositioning, pp,
        AM_SEEKING_NoPositioning);
  if State = psStopped then
    MediaControl.StopWhenReady;
end;

function TPlayer.Seekable: Boolean;
var
  cap: DWORD;
begin
  case SourceType of
    pstFile, pstURL:
      begin
        cap := AM_SEEKING_CanSeekAbsolute;
        Result := SUCCEEDED(MediaSeeking.CheckCapabilities(cap));
      end;
    pstDVD:
      Result := True;
  else
    Result := False;
  end;
end;

// FORMAT_WaveFormatExFFMPEG {35189950-CAC9-4C8D-819D-B6FAEE15DD9D}
// is not supported
procedure MediaTypeToMediaTypeInfo(const MediaType: AM_MEDIA_TYPE;
  var MediaTypeInfo: TMediaTypeInfo);
begin
  if (IsEqualGuid(MediaType.formattype, FORMAT_VideoInfo) or
      IsEqualGuid(MediaType.formattype, FORMAT_MPEGVideo)) and
      (MediaType.cbFormat >= SizeOf(TVideoInfoHeader)) then
    with PVideoInfoHeader(MediaType.pbFormat)^ do
    begin
      if AvgTimePerFrame > 0 then
        MediaTypeInfo.VideoFrameRate := 10000000 / AvgTimePerFrame;
      MediaTypeInfo.VideoWidth := bmiHeader.biWidth;
      MediaTypeInfo.VideoHeight := bmiHeader.biHeight;
      SetString(MediaTypeInfo.VideoCodec, PChar(@bmiHeader.biCompression),
          SizeOf(bmiHeader.biCompression));
    end
  else if (IsEqualGuid(MediaType.formattype, FORMAT_VideoInfo2) or
      IsEqualGuid(MediaType.formattype, FORMAT_MPEG2Video)) and
      (MediaType.cbFormat >= SizeOf(TVideoInfoHeader2)) then
    with PVideoInfoHeader2(MediaType.pbFormat)^ do
    begin
      if AvgTimePerFrame > 0 then
        MediaTypeInfo.VideoFrameRate := 10000000 / AvgTimePerFrame;
      MediaTypeInfo.VideoWidth := bmiHeader.biWidth;
      MediaTypeInfo.VideoHeight := bmiHeader.biHeight;
      SetString(MediaTypeInfo.VideoCodec, PChar(@bmiHeader.biCompression),
          SizeOf(bmiHeader.biCompression));
    end
  else if IsEqualGuid(MediaType.formattype, FORMAT_WaveFormatEx) and
      (MediaType.cbFormat >= SizeOf(TWaveFormatEx)) then
    with PWaveFormatEx(MediaType.pbFormat)^ do
    begin
      MediaTypeInfo.AudioSampleRate := nSamplesPerSec;
      MediaTypeInfo.AudioBitRate := nAvgBytesPerSec*8;
      MediaTypeInfo.AudioChannels := nChannels;
      MediaTypeInfo.AudioFormatTag := wFormatTag;
    end;
end;

procedure TPlayer.GetFilterInfo;
  procedure ClearMediaTypeInfo(var MediaTypeInfo: TMediaTypeInfo);
  begin
    with MediaTypeInfo do
    begin
      AudioSampleRate := 0;
      AudioBitRate := 0;
      AudioChannels := 0;
      AudioFormatTag := 0;
      VideoWidth := 0;
      VideoHeight := 0;
      VideoFrameRate := 0;
      VideoCodec := '';
    end;
  end;
var
  i: Integer;
  ef: IEnumFilters;
  info: FILTER_INFO;
  fetched: ULONG;
  ep: IEnumPins;
  issplitter: Boolean;
  pin: IPin;
  pindir: PIN_DIRECTION;
  mt: AM_MEDIA_TYPE;
  mti: TMediaTypeInfo;
begin
  SetLength(FilterInfo, 256);
  GraphBuilder.EnumFilters(ef);
  ClearMediaTypeInfo(MediaTypeInfo);
  ClearMediaTypeInfo(mti);
  i := 0;
  while ef.Next(1, FilterInfo[i].Filter, fetched) = S_OK do
    with FilterInfo[i] do
    begin
      Filter.QueryFilterInfo(info);
      info.pGraph := nil;
      Name := info.achName;
      if FAILED(Filter.QueryInterface(IID_ISpecifyPropertyPages, Prop)) then
        Prop := nil;

      issplitter := True;
      Filter.EnumPins(ep);
      while ep.Next(1, pin, fetched) = S_OK do
      begin
        pin.QueryDirection(pindir);
        if pindir = PINDIR_INPUT then
        begin
          pin.ConnectionMediaType(mt);
          if mt.pbFormat <> nil then
            CoTaskMemFree(mt.pbFormat);
          if not IsEqualGuid(mt.majortype, MEDIATYPE_Stream) then
          begin
            issplitter := False;
            break;
          end;
        end;

        if pindir = PINDIR_OUTPUT then
        begin
          pin.ConnectionMediaType(mt);
          MediaTypeToMediaTypeInfo(mt, mti);
          if mt.pbFormat <> nil then
            CoTaskMemFree(mt.pbFormat);
        end;
      end;

      if issplitter then
        MediaTypeInfo := mti;

      Inc(i);
      if i = Length(FilterInfo) then
        SetLength(FilterInfo, i + i div 4);
    end;
  SetLength(FilterInfo, i);
end;

procedure TPlayer.ShowFilterProperty(Index: Integer);
var
  pages: TCAGUID;
begin
  with FilterInfo[Index] do
  begin
    CheckResult(Prop.GetPages(pages));
    OleCreatePropertyFrame(Parent, 0, 0, PWideChar(WideString(Name)), 1,
        @Filter, pages.cElems, pages.pElems, 0, 0, nil);
    CoTaskMemFree(pages.pElems);
  end;
end;

function TPlayer.GetImageMF(var Buf: PChar): Cardinal;
var
  bi: TBitmapInfoHeader;
  nrc, bknrc: TMFVideoNormalizedRect;
  rc, bkrc: TRect;
  dib: PChar;
  dibsize: DWORD;
  ts: LONGLONG;
begin
  CheckResult(MFVideoDisplayControl.GetVideoPosition(bknrc, bkrc));
  nrc.left := 0;
  nrc.top := 0;
  nrc.right := 1;
  nrc.bottom := 1;
  rc.Left := 0;
  rc.Top := 0;
  rc.Right := VideoWidth;
  rc.Bottom := VideoHeight;
  CheckResult(MFVideoDisplayControl.SetVideoPosition(@nrc, @rc));
  bi.biSize := SizeOf(bi);
  CheckResult(MFVideoDisplayControl.GetCurrentImage(@bi, dib, dibsize, @ts));
  try
    CheckResult(MFVideoDisplayControl.SetVideoPosition(@bknrc, @bkrc));
    Result := SizeOf(TBitmapFileHeader) + SizeOf(bi) + dibsize;
    GetMem(Buf, Result);
    with PBitmapFileHeader(Buf)^ do
    begin
      bfType := Byte('B') + Byte('M') shl 8;
      bfSize := Result;
      bfReserved1 := 0;
      bfReserved2 := 0;
      bfOffBits := SizeOf(TBitmapFileHeader) + SizeOf(bi);
      Move(bi, Buf[SizeOf(TBitmapFileHeader)], SizeOf(bi));
      Move(dib^, Buf[bfOffBits], dibsize);
    end;
  finally
    CoTaskMemFree(dib);
  end;
end;

function TPlayer.GetImageBasicVideo(var Buf: PChar): Cardinal;
var
  bv: IBasicVideo;
  dibsize: Longint;
begin
  bv := GraphBuilder as IBasicVideo;
  CheckResult(bv.GetCurrentImage(dibsize, nil^));
  Result := SizeOf(TBitmapFileHeader) + dibsize;
  GetMem(Buf, Result);
  CheckResult(bv.GetCurrentImage(dibsize, Buf[SizeOf(TBitmapFileHeader)]));
  with PBitmapFileHeader(Buf)^ do
  begin
    bfType := Byte('B') + Byte('M') shl 8;
    bfSize := Result;
    bfReserved1 := 0;
    bfReserved2 := 0;
    bfOffBits := SizeOf(TBitmapFileHeader) +
        PBitmapInfoHeader(@Buf[SizeOf(TBitmapFileHeader)]).biSize;
  end;
end;

function TPlayer.GetImageSampleGrabber(var Buf: PChar): Cardinal;
var
  ba: IBasicAudio;
  pbi: PBitmapInfoHeader;
  bitssize, code: Longint;
  grabber: ISampleGrabber;
  grabberf: IBaseFilter;
  graph: IGraphBuilder;
  mt: AM_MEDIA_TYPE;
  startp, endp: LONGLONG;
begin
  CheckResult(CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC_SERVER,
      IID_IGraphBuilder, graph));
  CheckResult(CoCreateInstance(CLSID_SampleGrabber, nil, CLSCTX_INPROC_SERVER,
      IID_IBaseFilter, grabberf));
  grabber := grabberf as ISampleGrabber;
  FillChar(mt, SizeOf(mt), 0);
  mt.majortype := MEDIATYPE_Video;
  mt.subtype := MEDIASUBTYPE_RGB24;
  mt.formattype := GUID_NULL;
  grabber.SetMediaType(mt);
  grabber.SetBufferSamples(True);
  //grabber.SetOneShot(True);
  CheckResult(graph.AddFilter(grabberf, 'Sample Grabber'));

  graph.RenderFile(PWideChar(WideString(Path)), nil);
  MediaSeeking.GetCurrentPosition(endp);
  startp := endp - 10000*1000; // -1 sec.
  if startp < 0 then
    startp := 0;
  (graph as IMediaSeeking).SetPositions(
      startp, AM_SEEKING_AbsolutePositioning,
      endp, AM_SEEKING_AbsolutePositioning);
  (graph as IMediaFilter).SetSyncSource(nil);
  (graph as IVideoWindow).put_AutoShow(OAFALSE);
  ba := graph as IBasicAudio;
  if Assigned(ba) then
    ba.put_Volume(-10000);
  //MediaControl.StopWhenReady;
  //while MediaControl.GetState(10, fs) = VFW_S_STATE_INTERMEDIATE do;
  (graph as IMediaControl).Run;
  (graph as IMediaEvent).WaitForCompletion(1000, code);

  CheckResult(grabber.GetCurrentBuffer(bitssize, nil));
  CheckResult(grabber.GetConnectedMediaType(mt));
  try
    pbi := @PVideoInfoHeader(mt.pbFormat).bmiHeader;
    Result := SizeOf(TBitmapFileHeader) + pbi.biSize + Cardinal(bitssize);
    GetMem(Buf, Result);
    with PBitmapFileHeader(Buf)^ do
    begin
      bfType := Byte('B') + Byte('M') shl 8;
      bfSize := Result;
      bfReserved1 := 0;
      bfReserved2 := 0;
      bfOffBits := SizeOf(TBitmapFileHeader) + pbi.biSize;
      Move(pbi^, Buf[SizeOf(TBitmapFileHeader)], pbi.biSize);
      CheckResult(grabber.GetCurrentBuffer(bitssize, @Buf[bfOffBits]));
    end;
  finally
    CoTaskMemFree(mt.pbFormat);
  end;
end;

function TPlayer.Capture(var Buf: PChar): Cardinal;
begin
  if EVRFilter <> nil then
    Result := GetImageMF(Buf)
  else if VMR9Filter <> nil then
    Result := GetImageBasicVideo(Buf)
  else
    Result := GetImageSampleGrabber(Buf);
end;

{
function TPlayer.ClosedCaption: Boolean;
var
  state: AM_LINE21_CCSTATE;
begin
  Result := False;
  if not Assigned(Line21Decoder) then
    Exit;
  Line21Decoder.GetServiceState(state);
  Result := state = AM_L21_CCSTATE_On;
end;

procedure TPlayer.SetClosedCaption(Flag: Boolean);
begin
  if not Assigned(Line21Decoder) then
    Exit;
  if Flag then
    Line21Decoder.SetServiceState(AM_L21_CCSTATE_On)
  else
    Line21Decoder.SetServiceState(AM_L21_CCSTATE_Off);
end;  
}

initialization
  IsMultiThread := True;
  CoInitialize(nil);

finalization
  CoUninitialize;

end.
